# -*- coding: utf-8 -*-

import collections
import json
from typing import List, Dict, Optional, Iterable

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import format_url


class Subset(TM1Object):
    """ Abstraction of the TM1 Subset (dynamic and static)

    """

    def __init__(self, subset_name: str, dimension_name: str, hierarchy_name: str = None, alias: str = None,
                 expression: str = None, elements: Iterable[str] = None):
        """

        :param subset_name: String
        :param dimension_name: String
        :param hierarchy_name: String
        :param alias: String, alias that is active in this subset.
        :param expression: String
        :param elements: List, element names
        """
        self._dimension_name = dimension_name
        self._hierarchy_name = hierarchy_name if hierarchy_name else dimension_name
        self._subset_name = subset_name
        self._alias = alias
        self._expression = expression
        self._elements = list(elements) if elements else []

    @property
    def dimension_name(self) -> str:
        return self._dimension_name

    @dimension_name.setter
    def dimension_name(self, value: str):
        self._dimension_name = value

    @property
    def hierarchy_name(self) -> str:
        return self._hierarchy_name

    @hierarchy_name.setter
    def hierarchy_name(self, value: str):
        self._hierarchy_name = value

    @property
    def name(self) -> str:
        return self._subset_name

    @name.setter
    def name(self, value: str):
        self._subset_name = value

    @property
    def alias(self) -> str:
        return self._alias

    @alias.setter
    def alias(self, value: str):
        self._alias = value

    @property
    def expression(self) -> str:
        return self._expression

    @expression.setter
    def expression(self, value: str):
        self._expression = value

    @property
    def elements(self) -> List[str]:
        return self._elements

    @elements.setter
    def elements(self, value: List[str]):
        self._elements = value

    @property
    def type(self) -> str:
        if self.expression:
            return 'dynamic'
        return 'static'

    @property
    def is_dynamic(self) -> bool:
        return bool(self.expression)

    @property
    def is_static(self) -> bool:
        return not self.is_dynamic

    @classmethod
    def from_json(cls, subset_as_json: str) -> 'Subset':
        """ Alternative constructor
                :Parameters:
                    `subset_as_json` : string, JSON
                        representation of Subset as specified in CSDL

                :Returns:
                    `Subset` : an instance of this class
        """

        subset_as_dict = json.loads(subset_as_json)
        return cls.from_dict(subset_as_dict=subset_as_dict)

    @classmethod
    def from_dict(cls, subset_as_dict: Dict) -> 'Subset':
        return cls(dimension_name=subset_as_dict["UniqueName"][1:subset_as_dict["UniqueName"].find('].[')],
                   hierarchy_name=subset_as_dict.get("Hierarchy", {}).get("Name"),
                   subset_name=subset_as_dict['Name'],
                   alias=subset_as_dict.get('Alias'),
                   expression=subset_as_dict.get('Expression'),
                   elements=[element['Name'] for element in subset_as_dict.get('Elements', [])]
                   if not subset_as_dict.get('Expression') else None)

    @property
    def body(self) -> str:
        """ same logic here as in TM1 : when subset has expression its dynamic, otherwise static
        """
        return json.dumps(self.body_as_dict, ensure_ascii=False)

    @property
    def body_as_dict(self) -> Dict:
        """ same logic here as in TM1 : when subset has expression its dynamic, otherwise static
        """
        if self._expression:
            return self._construct_body_dynamic()
        else:
            return self._construct_body_static()

    def add_elements(self, elements: Iterable[str]):
        """ add Elements to static subsets
            :Parameters:
                `elements` : list of element names
        """
        self._elements = self._elements + list(elements)

    def _construct_body_dynamic(self) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self._subset_name
        if self.alias:
            body_as_dict['Alias'] = self._alias
        body_as_dict['Hierarchy@odata.bind'] = format_url(
            "Dimensions('{}')/Hierarchies('{}')",
            self._dimension_name,
            self._hierarchy_name)
        body_as_dict['Expression'] = self._expression
        return body_as_dict

    def _construct_body_static(self) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self._subset_name
        if self.alias:
            body_as_dict['Alias'] = self._alias
        body_as_dict['Hierarchy@odata.bind'] = format_url(
            "Dimensions('{}')/Hierarchies('{}')",
            self._dimension_name,
            self.hierarchy_name)
        if self.elements and len(self.elements) > 0:
            body_as_dict['Elements@odata.bind'] = [
                format_url("Dimensions('{}')/Hierarchies('{}')/Elements('{}')",
                           self.dimension_name, self.hierarchy_name, element)
                for element
                in self.elements]
        return body_as_dict


class AnonymousSubset(Subset):
    """ Abstraction of unregistered Subsets used in NativeViews (Check TM1py.ViewAxisSelection)

    """

    def __init__(self, dimension_name: str, hierarchy_name: Optional[str] = None, expression: Optional[str] = None,
                 elements: Optional[Iterable[str]] = None, alias: str = ''):
        Subset.__init__(self,
                        dimension_name=dimension_name,
                        hierarchy_name=hierarchy_name if hierarchy_name else dimension_name,
                        subset_name='',
                        alias=alias,
                        expression=expression,
                        elements=elements)

    @classmethod
    def from_json(cls, subset_as_json: str) -> 'Subset':
        """ Alternative constructor
                :Parameters:
                    `subset_as_json` : string, JSON
                        representation of Subset as specified in CSDL

                :Returns:
                    `Subset` : an instance of this class
        """
        subset_as_dict = json.loads(subset_as_json)
        return cls.from_dict(subset_as_dict=subset_as_dict)

    @classmethod
    def from_dict(cls, subset_as_dict: Dict) -> 'Subset':
        """Alternative constructor
        
        :param subset_as_dict: dictionary, representation of Subset as specified in CSDL
        :return: an instance of this class
        """
        return cls(dimension_name=subset_as_dict["Hierarchy"]["Dimension"]["Name"],
                   hierarchy_name=subset_as_dict["Hierarchy"]["Name"],
                   expression=subset_as_dict['Expression'],
                   alias=subset_as_dict['Alias'],
                   elements=[element['Name'] for element in subset_as_dict['Elements']]
                   if not subset_as_dict['Expression'] else None)

    def _construct_body_dynamic(self) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict['Hierarchy@odata.bind'] = format_url(
            "Dimensions('{}')/Hierarchies('{}')",
            self._dimension_name,
            self.hierarchy_name)
        if self.alias:
            body_as_dict['Alias'] = self._alias
        body_as_dict['Expression'] = self._expression
        return body_as_dict

    def _construct_body_static(self) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict['Hierarchy@odata.bind'] = format_url(
            "Dimensions('{}')/Hierarchies('{}')",
            self._dimension_name,
            self.hierarchy_name)
        if self.alias:
            body_as_dict['Alias'] = self._alias
        body_as_dict['Elements@odata.bind'] = [
            format_url(
                "Dimensions('{}')/Hierarchies('{}')/Elements('{}')",
                self.dimension_name,
                self.hierarchy_name,
                element)
            for element in self.elements]
        return body_as_dict



# -*- coding: utf-8 -*-
import json
from typing import Dict, Tuple, List, Optional

from requests import Response

from TM1py.Objects import Hierarchy, Element, ElementAttribute
from TM1py.Services.ElementService import ElementService
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Services.SubsetService import SubsetService
from TM1py.Utils.Utils import case_and_space_insensitive_equals, format_url, CaseAndSpaceInsensitiveDict, \
    CaseAndSpaceInsensitiveSet


class HierarchyService(ObjectService):
    """ Service to handle Object Updates for TM1 Hierarchies
    
    """

    # Tuple with TM1 Versions where Edges need to be created through TI, due to bug:
    # https://www.ibm.com/developerworks/community/forums/html/topic?id=75f2b99e-6961-4c71-9364-1d5e1e083eff
    EDGES_WORKAROUND_VERSIONS = ('11.0.002', '11.0.003', '11.1.000')

    def __init__(self, rest: RestService):
        super().__init__(rest)
        self.subsets = SubsetService(rest)
        self.elements = ElementService(rest)

    def create(self, hierarchy: Hierarchy, **kwargs):
        """ Create a hierarchy in an existing dimension

        :param hierarchy:
        :return:
        """
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies", hierarchy.dimension_name)
        response = self._rest.POST(url, hierarchy.body, **kwargs)
        return response

    def get(self, dimension_name: str, hierarchy_name: str, **kwargs) -> Hierarchy:
        """ get hierarchy

        :param dimension_name: name of the dimension
        :param hierarchy_name: name of the hierarchy
        :return:
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')?$expand=Edges,Elements,ElementAttributes,Subsets,DefaultMember",
            dimension_name,
            hierarchy_name)
        response = self._rest.GET(url, **kwargs)
        return Hierarchy.from_dict(response.json())

    def get_all_names(self, dimension_name: str, **kwargs) -> List[str]:
        """ get all names of existing Hierarchies in a dimension

        :param dimension_name:
        :return:
        """
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies?$select=Name", dimension_name)
        response = self._rest.GET(url, **kwargs)
        return [hierarchy["Name"] for hierarchy in response.json()["value"]]

    def update(self, hierarchy: Hierarchy, **kwargs) -> List[Response]:
        """ update a hierarchy. It's a two step process: 
        1. Update Hierarchy
        2. Update Element-Attributes

        Function caters for Bug with Edge Creation:
        https://www.ibm.com/developerworks/community/forums/html/topic?id=75f2b99e-6961-4c71-9364-1d5e1e083eff

        :param hierarchy: instance of TM1py.Hierarchy
        :return: list of responses
        """
        # functions returns multiple responses
        responses = list()
        # 1. Update Hierarchy
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')", hierarchy.dimension_name, hierarchy.name)
        # Workaround EDGES: Handle Issue, that Edges cant be created in one batch with the Hierarchy in certain versions
        hierarchy_body = hierarchy.body_as_dict
        if self.version[0:8] in self.EDGES_WORKAROUND_VERSIONS:
            del hierarchy_body["Edges"]
        responses.append(self._rest.PATCH(url, json.dumps(hierarchy_body), **kwargs))

        # 2. Update Attributes
        responses.append(self.update_element_attributes(hierarchy=hierarchy, **kwargs))

        # Workaround EDGES
        if self.version[0:8] in self.EDGES_WORKAROUND_VERSIONS:
            from TM1py.Services import ProcessService
            process_service = ProcessService(self._rest)
            ti_function = "HierarchyElementComponentAdd('{}', '{}', '{}', '{}', {});"
            ti_statements = [ti_function.format(hierarchy.dimension_name, hierarchy.name,
                                                edge[0],
                                                edge[1],
                                                hierarchy.edges[(edge[0], edge[1])])
                             for edge
                             in hierarchy.edges]
            responses.append(process_service.execute_ti_code(lines_prolog=ti_statements, **kwargs))

        return responses

    def update_or_create(self, hierarchy: Hierarchy, **kwargs):
        """ update if exists else create

        :param Hierarchy:
        :return:
        """
        if self.exists(dimension_name=hierarchy.dimension_name, hierarchy_name=hierarchy.name, **kwargs):
            self.update(hierarchy=hierarchy, **kwargs)
        else:
            self.create(hierarchy=hierarchy, **kwargs)

    def exists(self, dimension_name: str, hierarchy_name: str, **kwargs) -> bool:
        """

        :param dimension_name: 
        :param hierarchy_name: 
        :return: 
        """
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')", dimension_name, hierarchy_name)
        return self._exists(url, **kwargs)

    def delete(self, dimension_name: str, hierarchy_name: str, **kwargs) -> Response:
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')", dimension_name, hierarchy_name)
        return self._rest.DELETE(url, **kwargs)

    def get_hierarchy_summary(self, dimension_name: str, hierarchy_name: str, **kwargs) -> Dict[str, int]:
        hierarchy_properties = ("Elements", "Edges", "ElementAttributes", "Members", "Levels")
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')?$expand=Edges/$count,Elements/$count,"
            "ElementAttributes/$count,Members/$count,Levels/$count&$select=Cardinality",
            dimension_name,
            hierarchy_name)
        hierary_summary_raw = self._rest.GET(url, **kwargs).json()

        return {hierarchy_property: hierary_summary_raw[hierarchy_property + "@odata.count"]
                for hierarchy_property
                in hierarchy_properties}

    def update_element_attributes(self, hierarchy: Hierarchy, **kwargs):
        """ Update the elementattributes of a hierarchy

        :param hierarchy: Instance of TM1py.Hierarchy
        :return:
        """
        # get existing attributes first
        existing_element_attributes = self.elements.get_element_attributes(
            dimension_name=hierarchy.dimension_name,
            hierarchy_name=hierarchy.name,
            **kwargs)
        existing_element_attributes = CaseAndSpaceInsensitiveDict({ea.name: ea for ea in existing_element_attributes})

        attributes_to_create = list()
        attributes_to_delete = list()
        attributes_to_update = list()

        for element_attribute in hierarchy.element_attributes:
            if element_attribute.name not in existing_element_attributes:
                attributes_to_create.append(element_attribute)
                continue

            existing_element_attribute = existing_element_attributes[element_attribute.name]
            if not existing_element_attribute.attribute_type == element_attribute.attribute_type:
                attributes_to_update.append(element_attribute)
                continue

        for existing_element_attribute in existing_element_attributes:
            if existing_element_attribute not in CaseAndSpaceInsensitiveSet(
                    [ea.name for ea in hierarchy.element_attributes]):
                attributes_to_delete.append(existing_element_attribute)

        for element_attribute in attributes_to_create:
            self.elements.create_element_attribute(
                dimension_name=hierarchy.dimension_name,
                hierarchy_name=hierarchy.name,
                element_attribute=element_attribute,
                **kwargs)

        for element_attribute in attributes_to_delete:
            self.elements.delete_element_attribute(
                dimension_name=hierarchy.dimension_name,
                hierarchy_name=hierarchy.name,
                element_attribute=element_attribute,
                **kwargs)

        for element_attribute in attributes_to_update:
            self.elements.delete_element_attribute(
                dimension_name=hierarchy.dimension_name,
                hierarchy_name=hierarchy.name,
                element_attribute=element_attribute.name,
                **kwargs)
            self.elements.create_element_attribute(
                dimension_name=hierarchy.dimension_name,
                hierarchy_name=hierarchy.name,
                element_attribute=element_attribute,
                **kwargs)

    def get_default_member(self, dimension_name: str, hierarchy_name: str = None, **kwargs) -> Optional[str]:
        """ Get the defined default_member for a Hierarchy.
        Will return the element with index 1, if default member is not specified explicitly in }HierarchyProperty Cube

        :param dimension_name:
        :param hierarchy_name:
        :return: String, name of Member
        """
        url = format_url(
            "/api/v1/Dimensions('{dimension}')/Hierarchies('{hierarchy}')/DefaultMember",
            dimension=dimension_name,
            hierarchy=hierarchy_name if hierarchy_name else dimension_name)
        response = self._rest.GET(url=url, **kwargs)

        if not response.text:
            return None
        return response.json()["Name"]

    def update_default_member(self, dimension_name: str, hierarchy_name: str = None, member_name: str = "",
                              **kwargs) -> Response:
        """ Update the default member of a hierarchy.
        Currently implemented through TI, since TM1 API does not supports default member updates yet.

        :param dimension_name:
        :param hierarchy_name:
        :param member_name:
        :return:
        """
        from TM1py import ProcessService, CellService
        if hierarchy_name and not case_and_space_insensitive_equals(dimension_name, hierarchy_name):
            dimension = "{}:{}".format(dimension_name, hierarchy_name)
        else:
            dimension = dimension_name
        cells = {(dimension, 'hierarchy0', 'defaultMember'): member_name}

        CellService(self._rest).write_values(
            cube_name="}HierarchyProperties",
            cellset_as_dict=cells,
            dimensions=('}Dimensions', '}Hierarchies', '}HierarchyProperties'),
            **kwargs)

        return ProcessService(self._rest).execute_ti_code(
            lines_prolog=format_url("RefreshMdxHierarchy('{}');", dimension_name),
            **kwargs)

    def remove_all_edges(self, dimension_name: str, hierarchy_name: str = None, **kwargs) -> Response:
        if not hierarchy_name:
            hierarchy_name = dimension_name
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')", dimension_name, hierarchy_name)
        body = {
            "Edges": []
        }
        return self._rest.PATCH(url=url, data=json.dumps(body), **kwargs)

    def remove_edges_under_consolidation(self, dimension_name: str, hierarchy_name: str,
                                         consolidation_element: str, **kwargs) -> List[Response]:
        """
        :param dimension_name: Name of the dimension
        :param hierarchy_name: Name of the hierarchy
        :param consolidation_element: Name of the Consolidated element
        :return: response
        """
        hierarchy = self.get(dimension_name, hierarchy_name)
        from TM1py.Services import ElementService
        element_service = ElementService(self._rest)
        elements_under_consolidations = element_service.get_members_under_consolidation(dimension_name, hierarchy_name,
                                                                                        consolidation_element)
        elements_under_consolidations.append(consolidation_element)
        remove_edges = []
        for (parent, component) in hierarchy.edges:
            if parent in elements_under_consolidations and component in elements_under_consolidations:
                remove_edges.append((parent, component))
        hierarchy.remove_edges(remove_edges)
        return self.update(hierarchy, **kwargs)

    def add_edges(self, dimension_name: str, hierarchy_name: str = None, edges: Dict[Tuple[str, str], int] = None,
                  **kwargs) -> Response:
        """ Add Edges to hierarchy. Fails if one edge already exists.

        :param dimension_name:
        :param hierarchy_name:
        :param edges:
        :return:
        """
        return self.elements.add_edges(dimension_name, hierarchy_name, edges, **kwargs)

    def add_elements(self, dimension_name: str, hierarchy_name: str, elements: List[Element], **kwargs):
        """ Add elements to hierarchy. Fails if one element already exists.

        :param dimension_name:
        :param hierarchy_name:
        :param elements:
        :return:
        """
        return self.elements.add_elements(dimension_name, hierarchy_name, elements, **kwargs)

    def add_element_attributes(self, dimension_name: str, hierarchy_name: str,
                               element_attributes: List[ElementAttribute], **kwargs):
        """ Add element attributes to hierarchy. Fails if one element attribute already exists.

        :param dimension_name:
        :param hierarchy_name:
        :param element_attributes:
        :return:
        """
        return self.elements.add_element_attributes(dimension_name, hierarchy_name, element_attributes, **kwargs)

    def is_balanced(self, dimension_name: str, hierarchy_name: str, **kwargs):
        """ Check if hierarchy is balanced

        :param dimension_name:
        :param hierarchy_name:
        :return:
        """
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/Structure/$value",
            dimension_name,
            hierarchy_name)
        structure = int(self._rest.GET(url, **kwargs).text)
        # 0 = balanced, 2 = unbalanced
        if structure == 0:
            return True
        elif structure == 2:
            return False
        else:
            raise RuntimeError(f"Unexpected return value from TM1 API request: {str(structure)}")



import pickle

from TM1py.Services import HierarchyService, SecurityService, ApplicationService, SubsetService, ServerService, \
    MonitoringService, ProcessService, PowerBiService, AnnotationService, ViewService, RestService, CellService, \
    ChoreService, DimensionService, CubeService, ElementService, SandboxService, GitService


class TM1Service:
    """ All features of TM1py are exposed through this service
    
    Can be saved and restored from File, to avoid multiple authentication with TM1.
    """

    def __init__(self, **kwargs):
        self._tm1_rest = RestService(**kwargs)
        self.annotations = AnnotationService(self._tm1_rest)
        self.cells = CellService(self._tm1_rest)
        self.chores = ChoreService(self._tm1_rest)
        self.cubes = CubeService(self._tm1_rest)
        self.dimensions = DimensionService(self._tm1_rest)
        self.elements = ElementService(self._tm1_rest)
        self.git = GitService(self._tm1_rest)
        self.hierarchies = HierarchyService(self._tm1_rest)
        self.monitoring = MonitoringService(self._tm1_rest)
        self.power_bi = PowerBiService(self._tm1_rest)
        self.processes = ProcessService(self._tm1_rest)
        self.security = SecurityService(self._tm1_rest)
        self.server = ServerService(self._tm1_rest)
        self.subsets = SubsetService(self._tm1_rest)
        self.applications = ApplicationService(self._tm1_rest)
        self.views = ViewService(self._tm1_rest)
        self.sandboxes = SandboxService(self._tm1_rest)
        self.git = GitService(self._tm1_rest)

    def logout(self, **kwargs):
        self._tm1_rest.logout(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.logout()

    @property
    def whoami(self):
        return self.security.get_current_user()

    @property
    def version(self):
        return self._tm1_rest.version

    @property
    def connection(self):
        return self._tm1_rest

    def save_to_file(self, file_name):
        with open(file_name, 'wb') as file:
            pickle.dump(self, file)

    @classmethod
    def restore_from_file(cls, file_name):
        with open(file_name, 'rb') as file:
            return pickle.load(file)

    def _instantiate_services(self):
        self.annotations = AnnotationService(self._tm1_rest)
        self.cells = CellService(self._tm1_rest)
        self.chores = ChoreService(self._tm1_rest)
        self.cubes = CubeService(self._tm1_rest)
        self.dimensions = DimensionService(self._tm1_rest)
        self.elements = ElementService(self._tm1_rest)
        self.git = GitService(self._tm1_rest)
        self.hierarchies = HierarchyService(self._tm1_rest)
        self.monitoring = MonitoringService(self._tm1_rest)
        self.power_bi = PowerBiService(self._tm1_rest)
        self.processes = ProcessService(self._tm1_rest)
        self.security = SecurityService(self._tm1_rest)
        self.server = ServerService(self._tm1_rest)
        self.subsets = SubsetService(self._tm1_rest)
        self.applications = ApplicationService(self._tm1_rest)
        self.views = ViewService(self._tm1_rest)
        self.sandboxes = SandboxService(self._tm1_rest)
        self.git = GitService(self._tm1_rest)

    def re_authenticate(self):
        self._tm1_rest = RestService(**self.connection._kwargs)
        self._instantiate_services()



# -*- coding: utf-8 -*-

from TM1py.Exceptions import TM1pyRestException
from TM1py.Services import RestService
from TM1py.Utils import format_url


class ObjectService:
    """ Parent class for all Object Services
    
    """

    ELEMENT_ATTRIBUTES_PREFIX = "}ElementAttributes_"
    SANDBOX_DIMENSION = "Sandboxes"

    BINARY_HTTP_HEADER = {'Content-Type': 'application/octet-stream; odata.streaming=true'}

    def __init__(self, rest_service: RestService):
        """ Constructor, Create an instance of ObjectService
        
        :param rest_service: 
        """
        self._rest = rest_service

    def determine_actual_object_name(self, object_class: str, object_name: str, **kwargs) -> str:
        url = format_url(
            "/api/v1/{}?$filter=tolower(replace(Name, ' ', '')) eq '{}'",
            object_class,
            object_name.replace(" ", "").lower())
        response = self._rest.GET(url, **kwargs)

        if len(response.json()["value"]) == 0:
            raise ValueError("Object '{}' of type '{}' doesn't exist".format(object_name, object_class))

        return response.json()["value"][0]["Name"]

    def _exists(self, url: str, **kwargs) -> bool:
        """ Check if resource exists in the TM1 Server
        
        :param url: 
        :return: 
        """
        try:
            self._rest.GET(url, **kwargs)
            return True
        except TM1pyRestException as e:
            if e.status_code == 404:
                return False
            raise e

    @property
    def version(self) -> str:
        return self._rest.version

    @property
    def is_admin(self) -> bool:
        return self._rest.is_admin



# -*- coding: utf-8 -*-
from typing import List

from TM1py.Objects.TM1Object import TM1Object


class Rules(TM1Object):
    """
        Abstraction of Rules on a cube.

        rules_analytics
            A collection of rulestatements, where each statement is stored in uppercase without linebreaks.
            comments are not included.

    """
    KEYWORDS = ['SKIPCHECK', 'FEEDSTRINGS', 'UNDEFVALS', 'FEEDERS']

    def __init__(self, rules: str):
        self._text = rules
        self._rules_analytics = []
        self.init_analytics()

    # self._rules_analytics_upper serves for analysis on cube rules
    def init_analytics(self):
        text_without_comments = '\n'.join(
            [rule
             for rule in self._text.split('\n')
             if len(rule.strip()) > 0 and rule.strip()[0] != '#'])
        for statement in text_without_comments.split(';'):
            if len(statement.strip()) > 0:
                self._rules_analytics.append(statement.replace('\n', '').upper())

    @property
    def text(self) -> str:
        return self._text

    @property
    def rules_analytics(self) -> List[str]:
        return self._rules_analytics

    @property
    def rule_statements(self) -> List[str]:
        if self.has_feeders:
            return self.rules_analytics[:self._rules_analytics.index('FEEDERS')]
        return self.rules_analytics

    @property
    def feeder_statements(self) -> List[str]:
        if self.has_feeders:
            return self.rules_analytics[self._rules_analytics.index('FEEDERS') + 1:]
        return []

    @property
    def skipcheck(self) -> bool:
        for rule in self._rules_analytics[0:5]:
            if rule == 'SKIPCHECK':
                return True
        return False

    @property
    def undefvals(self) -> bool:
        for rule in self._rules_analytics[0:5]:
            if rule == 'UNDEFVALS':
                return True
        return False

    @property
    def feedstrings(self) -> bool:
        for rule in self._rules_analytics[0:5]:
            if rule == 'FEEDSTRINGS':
                return True
        return False

    @property
    def has_feeders(self) -> bool:
        if 'FEEDERS' in self._rules_analytics:
            # has feeders declaration
            feeders = self.rules_analytics[self._rules_analytics.index('FEEDERS'):]
            # has at least one actual feeder statements
            return len(feeders) > 1
        return False

    def __len__(self):
        return len(self.rules_analytics)

    # iterate through actual rule statments without linebreaks. Ignore comments.
    def __iter__(self):
        return iter(self.rules_analytics)

    def __str__(self):
        return self.text



import collections
import csv
import functools
import http.client as http_client
import json
import math
import re
import ssl
import urllib.parse as urlparse
from contextlib import suppress
from enum import Enum, unique
from io import StringIO
from typing import Any, Dict, List, Tuple, Iterable, Optional, Generator, Union

import requests
from requests.adapters import HTTPAdapter
from mdxpy import MdxBuilder, Member

from TM1py.Exceptions.Exceptions import TM1pyVersionException, TM1pyNotAdminException

try:
    import pandas as pd
    import numpy as np

    _has_pandas = True
except ImportError:
    _has_pandas = False


def require_admin(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.is_admin:
            raise TM1pyNotAdminException(func.__name__)
        return func(self, *args, **kwargs)

    return wrapper


def require_version(version):
    """ Higher order function to check required version for TM1py function
    """

    def wrap(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if not verify_version(required_version=version, version=self.version):
                raise TM1pyVersionException(func.__name__, version)
            return func(self, *args, **kwargs)

        return wrapper

    return wrap


def require_pandas(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            import pandas
            return func(self, *args, **kwargs)
        except ImportError:
            raise ImportError(f"Function '{func.__name__}' requires pandas")

    return wrapper


def get_all_servers_from_adminhost(adminhost='localhost', port=None, use_ssl=False) -> List:
    from TM1py.Objects import Server
    """ Ask Adminhost for TM1 Servers
    :param adminhost: IP or DNS Alias of the adminhost
    :param port: numeric port to connect to adminhost
    :param ssl: True for secure connection
    :return: List of Servers (instances of the TM1py.Server class)
    """

    if not use_ssl:
        conn = http_client.HTTPConnection(adminhost, port or 5895)
    else:
        conn = http_client.HTTPSConnection(adminhost, port or 5898, context=ssl._create_unverified_context())
    request = '/api/v1/Servers'
    conn.request('GET', request, body='')
    response = conn.getresponse().read().decode('utf-8')
    response_as_dict = json.loads(response)
    servers = []
    for server_as_dict in response_as_dict['value']:
        server = Server(server_as_dict)
        servers.append(server)
    return servers


def create_server_on_adminhost(adminhost: str = 'localhost', server_as_dict: Dict = None):
    from TM1py.Objects import Server
    """  Create new TM1 instance on Adminhost
    :param adminhost: IP or DNS Alias of the adminhost
    :param server_as_dict: 
            server_as_dict = {
                "Name":"MyModel1",
                "IPAddress":"172.20.10.10",
                "IPv6Address":None,
                "PortNumber":12345,
                "UsingSSL": True,
                "ClientMessagePortNumber":61098,
                "HTTPPortNumber":12999,
                "ClientExportSSLSvrCert":True,
                "ClientExportSSLSvrKeyID":"whateverExportSSLSvrKeyID",
                "AcceptingClients":True }
    :return: instance of TM1py.Server
    """

    if not server_as_dict:
        raise ValueError("server_as_dict must be provided")

    if not adminhost:
        adminhost = 'localhost'

    url = f"http://{adminhost}:5895/api/v1/Servers"
    response = requests.post(url, data=json.dumps(server_as_dict), headers={'Content-Type': 'application/json'})
    response.raise_for_status()

    return Server(response.json())


def delete_server_on_adminhost(adminhost: str = None, server_name: str = None):
    if not server_name:
        raise ValueError("server_name must be provided")

    if not adminhost:
        adminhost = 'localhost'

    url = f"http://{adminhost}:5895/api/v1/Servers('{server_name}')"
    response = requests.delete(url, headers={'Content-Type': 'application/json'})
    response.raise_for_status()


def update_server_on_adminhost(adminhost: str = 'localhost', server_as_dict: Dict = None):
    from TM1py.Objects import Server
    """  Create new TM1 instance on Adminhost
    :param adminhost: IP or DNS Alias of the adminhost
    :param server_as_dict: 
            server_as_dict = {
                "Name":"MyModel1",
                "IPAddress":"172.20.10.10",
                "IPv6Address":None,
                "PortNumber":12345,
                "UsingSSL": True,
                "ClientMessagePortNumber":61098,
                "HTTPPortNumber":12999,
                "ClientExportSSLSvrCert":True,
                "ClientExportSSLSvrKeyID":"whateverExportSSLSvrKeyID",
                "AcceptingClients":True }
    :return: instance of TM1py.Server
    """

    if not server_as_dict:
        raise ValueError("server_as_dict must be provided")

    if not adminhost:
        adminhost = 'localhost'

    url = f"http://{adminhost}:5895/api/v1/Servers"
    response = requests.patch(url, body=json.dumps(server_as_dict), headers={'Content-Type': 'application/json'})
    response.raise_for_status()

    return Server(response.json())


def build_url_friendly_object_name(object_name: str) -> str:
    return object_name.replace("'", "''").replace('%', '%25').replace('#', '%23')


def format_url(url, *args: str, **kwargs: str) -> str:
    """ build url and escape single quotes in args and kwargs
    :param url: url with {} placeholders
    :param args: arguments to placeholders
    :return:
    """
    args = [build_url_friendly_object_name(arg) if isinstance(arg, str) else arg
            for arg
            in args]

    kwargs = {key: build_url_friendly_object_name(value) if isinstance(value, str) else value
              for key, value
              in kwargs.items()}

    return url.format(*args, **kwargs)


def abbreviate_mdx(mdx: str, size=100) -> str:
    if len(mdx) < size:
        return mdx
    else:
        return mdx[:size] + "..."


def integerize_version(version: str) -> int:
    return int(version[0:4].replace(".", ""))


def verify_version(required_version: str, version: str) -> bool:
    return integerize_version(version) >= integerize_version(required_version)


def case_and_space_insensitive_equals(item1: str, item2: str) -> bool:
    return lower_and_drop_spaces(item1) == lower_and_drop_spaces(item2)


def extract_axes_from_cellset(raw_cellset_as_dict: Dict) -> Tuple[Any, ...]:
    raw_axes = raw_cellset_as_dict['Axes']

    axes = list()

    for axis in raw_axes:
        if axis and 'Tuples' in axis and len(axis['Tuples']) > 0:
            axes.append(axis)

    return tuple(axes)


def extract_unique_names_from_members(members: Iterable[Dict]) -> List[str]:
    """ Extract list of unique names from part of the cellset response
    in:
    [{'UniqueName': '[dim1].[dim1].[elem1]', 'Element': {'UniqueName': '[dim1].[dim1].[elem1]'}},
    {'UniqueName': '[dim2].[dim2].[elem3]', 'Element': {'UniqueName': '[dim2].[dim2].[elem3]'}}]
    out:
    ["[dim1].[dim1].[elem1]", "[dim2].[dim2].[elem3]"]
    :param members: dictionary
    :return: list of unique names
    """
    return [m['Element']['UniqueName'] if 'Element' in m and m['Element'] else m['UniqueName']
            for m
            in members]


def extract_element_names_from_members(members: Iterable[Dict]) -> List[str]:
    """ Extract list of unique names from part of the cellset response
    in:
    [{'UniqueName': '[dim1].[dim1].[elem1]', 'Element': {'UniqueName': '[dim1].[dim1].[elem1]'}},
    {'UniqueName': '[dim2].[dim2].[elem3]', 'Element': {'UniqueName': '[dim2].[dim2].[elem3]'}}]
    out:
    ["elem1", "elem3"]
    :param members: dictionary
    :return: list of unique names
    """
    return [m['Element']['Name'] if 'Element' in m and m['Element'] else m['Name']
            for m
            in members]


def sort_coordinates(cube_dimensions: Iterable[str], unsorted_coordinates: Iterable[str],
                     element_unique_names: True) -> Tuple[str]:
    sorted_coordinates = []
    for dimension in cube_dimensions:
        # could be more than one hierarchy!
        address_elements = [item for item in unsorted_coordinates if item.startswith('[' + dimension + '].')]
        # address_elements could be ( [dim1].[hier1].[elem1], [dim1].[hier2].[elem3] )
        for address_element in address_elements:
            if element_unique_names:
                coordinate = address_element
            else:
                coordinate = element_name_from_element_unique_name(address_element)
            sorted_coordinates.append(coordinate)
    return tuple(sorted_coordinates)


def build_content_from_cellset_dict(
        raw_cellset_as_dict: Dict,
        top: Optional[int] = None,
        element_unique_names: bool = True,
        skip_cell_properties: bool = False) -> 'CaseAndSpaceInsensitiveTuplesDict':
    """ transform raw cellset data into concise dictionary
    :param raw_cellset_as_dict:
    :param top: Int, number of cells to return (counting from top)
    :param element_unique_names: '[d1].[h1].[e1]' or 'e1'
    :param skip_cell_properties: cell values in result dictionary, instead of cell_properties dictionary
    :return:
    """
    cube_dimensions = [dim['Name'] for dim in raw_cellset_as_dict['Cube']['Dimensions']]

    cells = raw_cellset_as_dict['Cells']
    axes = extract_axes_from_cellset(raw_cellset_as_dict=raw_cellset_as_dict)

    content_as_dict = CaseAndSpaceInsensitiveTuplesDict()
    for cell_ordinal, cell in enumerate(cells[:top or len(cells)]):
        # if skip is used in execution we must use the original ordinal from the cell, if not we can simply enumerate
        cell_ordinal = cell.get("Ordinal", cell_ordinal)

        coordinates = []
        for axis_ordinal, axis in enumerate(axes):

            if axis_ordinal == 0:
                index_columns = cell_ordinal % axis['Cardinality']
                coordinate = extract_unique_names_from_members(axis['Tuples'][index_columns]['Members'])
                coordinates.extend(coordinate)

            else:
                tuple_ordinal = cell_ordinal
                for pre_axis_ordinal in range(axis_ordinal):
                    tuple_ordinal = tuple_ordinal // axes[pre_axis_ordinal]['Cardinality']

                tuple_ordinal = tuple_ordinal % axis.get('Cardinality')
                coordinate = extract_unique_names_from_members(axis['Tuples'][tuple_ordinal]['Members'])
                coordinates.extend(coordinate)

        coordinates = sort_coordinates(cube_dimensions, coordinates, element_unique_names=element_unique_names)
        content_as_dict[coordinates] = cell['Value'] if skip_cell_properties else cell
    return content_as_dict


def _build_headers_for_csv(row_axis: Dict, column_axis: Dict, row_dimensions: List[str], column_dimensions: List[str],
                           include_attributes: bool):
    if not include_attributes:
        return [dimension_name_from_element_unique_name(dimension)
                for dimension
                in row_dimensions + column_dimensions] + ['Value']

    headers = list()
    if row_axis:
        members = row_axis["Tuples"][0]['Members']
        for dimension, member in zip(row_dimensions, members):
            headers.append(dimension_name_from_element_unique_name(dimension))
            for attribute in member['Attributes']:
                headers.append(attribute)

    members = column_axis["Tuples"][0]['Members']
    for dimension, member in zip(column_dimensions, members):
        headers.append(dimension_name_from_element_unique_name(dimension))
        for attribute in member['Attributes']:
            headers.append(attribute)

    return headers + ['Value']


def build_csv_from_cellset_dict(
        row_dimensions: List[str],
        column_dimensions: List[str],
        raw_cellset_as_dict: Dict,
        top: Optional[int] = None,
        csv_dialect: 'csv.Dialect' = None,
        line_separator: str = "\r\n",
        value_separator: str = ",",
        include_attributes: bool = False,
        include_headers: bool = True) -> str:
    """ transform raw cellset data into concise dictionary
    :param column_dimensions:
    :param row_dimensions:
    :param raw_cellset_as_dict:
    :param top: Maximum Number of cells
    :param csv_dialect: provide all csv output settings through standard library csv.Dialect
        If not provided dialect is created based on line_separator and value_separator arguments.
    :param line_separator:
    :param value_separator:
    :param include_attributes: include attribute columns
    :param include_headers: bool
    :return:
    """

    cells = raw_cellset_as_dict['Cells']
    # empty cellsets produce "" in order to be compliant with previous implementation that used `/Content` API endpoint
    if len(cells) == 0:
        return ""

    if csv_dialect is None:
        csv.register_dialect("TM1py", delimiter=value_separator, lineterminator=line_separator)
        csv_dialect = csv.get_dialect("TM1py")

    csv_content = StringIO()
    csv_writer = csv.writer(csv_content, dialect=csv_dialect)

    axes = extract_axes_from_cellset(raw_cellset_as_dict=raw_cellset_as_dict)
    column_axis = axes[0]
    if len(axes) > 1:
        row_axis = axes[1]
    else:
        row_axis = list()

    num_headers = 0
    if include_headers:
        headers = _build_headers_for_csv(row_axis, column_axis, row_dimensions, column_dimensions, include_attributes)
        csv_writer.writerow(headers)
        num_headers = len(headers)

    for ordinal, cell in enumerate(cells[:top or len(cells)]):
        # if skip is used in execution we must use the original ordinal from the cell, if not we can simply enumerate
        ordinal = cell.get("Ordinal", ordinal)

        line = []
        if column_axis and row_axis:
            index_rows = ordinal // column_axis['Cardinality'] % row_axis['Cardinality']
            index_columns = ordinal % column_axis['Cardinality']

            line_items = _build_csv_line_items_from_axis_tuple(
                members=row_axis['Tuples'][index_rows]['Members'],
                include_attributes=include_attributes)
            line.extend(line_items)

            line_items = _build_csv_line_items_from_axis_tuple(
                members=column_axis['Tuples'][index_columns]['Members'],
                include_attributes=include_attributes)
            line.extend(line_items)

        elif column_axis:
            index_rows = ordinal % column_axis['Cardinality']

            line_items = _build_csv_line_items_from_axis_tuple(
                members=column_axis['Tuples'][index_rows]['Members'],
                include_attributes=include_attributes)
            line.extend(line_items)

        line.append(str(cell["Value"] or ""))
        if include_attributes and include_headers and not len(line) == num_headers:
            raise ValueError("Invalid response. With 'include_attributes' as True,"
                             " Attributes must be requested explicitly as PROPERTIES in the MDX")
        csv_writer.writerow(line)

    return csv_content.getvalue().strip()


def _build_csv_line_items_from_axis_tuple(members: Dict, include_attributes: bool = False) -> List[str]:
    if not include_attributes:
        return extract_element_names_from_members(members)

    else:
        line_items = list()
        for member in members:

            element_name = member['Element']['Name'] if 'Element' in member and member['Element'] else member['Name']
            line_items.append(element_name)

            attribute_values = list(member['Attributes'].values())
            for attribute_value in attribute_values:
                line_items.append(str(attribute_value) if attribute_value else '')

        return line_items


def build_ui_arrays_from_cellset(raw_cellset_as_dict: Dict, value_precision: int):
    """ Transform raw 1,2 or 3-dimension cellset data into concise dictionary
    * Useful for grids or charting libraries that want an array of cell values per row
    * Returns 3-dimensional cell structure for tabbed grids or multiple charts
    * Rows and pages are dicts, addressable by their name. Proper order of rows can be obtained in headers[1]
    * Example 'cells' return format:
        'cells': {
            '10100': {
                'Net Operating Income': [ 19832724.72429739,
                                          20365654.788303416,
                                          20729201.329183243,
                                          20480205.20121749],
                'Revenue': [ 28981046.50724231,
                             29512482.207418434,
                             29913730.038971487,
                             29563345.9542385]},
            '10200': {
                'Net Operating Income': [ 9853293.623709997,
                                           10277650.763958748,
                                           10466934.096533755,
                                           10333095.839474997],
                'Revenue': [ 13888143.710000003,
                             14300216.43,
                             14502421.63,
                             14321501.940000001]}
        },
    :param raw_cellset_as_dict: raw data from TM1
    :param value_precision: Integer (optional) specifying number of decimal places to return
    :return: dict : { titles: [], headers: [axis][], cells: { Page0: { Row0: { [row values], Row1: [], ...}, ...}, ...} }
    """
    header_map = build_headers_from_cellset(raw_cellset_as_dict, force_header_dimensionality=3)
    titles = header_map['titles']
    headers = header_map['headers']
    cardinality = header_map['cardinality']
    value_format_string = ""

    if value_precision:
        value_format_string = "{{0:.{}f}}".format(value_precision)

    cells = {}
    ordinal_cells = 0
    for z in range(cardinality[2]):
        z_header = headers[2][z]['name']
        pages = {}
        for y in range(cardinality[1]):
            y_header = headers[1][y]['name']
            row = []
            for x in range(cardinality[0]):
                raw_value = raw_cellset_as_dict['Cells'][ordinal_cells]['Value'] or 0
                if value_precision:
                    row.append(float(value_format_string.format(raw_value)))
                else:
                    row.append(raw_value)
                ordinal_cells += 1
            pages[y_header] = row
        cells[z_header] = pages
    return {'titles': titles, 'headers': headers, 'cells': cells}


def build_ui_dygraph_arrays_from_cellset(raw_cellset_as_dict: Dict, value_precision: int = None):
    """ Transform raw 1,2 or 3-dimension cellset data into dygraph-friendly format
    * Useful for grids or charting libraries that want an array of cell values per column
    * Returns 3-dimensional cell structure for tabbed grids or multiple charts
    * Example 'cells' return format:
        'cells': {
            '10100': [
                ['Q1-2004', 28981046.50724231, 19832724.72429739],
                ['Q2-2004', 29512482.207418434, 20365654.788303416],
                ['Q3-2004', 29913730.038971487, 20729201.329183243],
                ['Q4-2004', 29563345.9542385, 20480205.20121749]],
            '10200': [
                ['Q1-2004', 13888143.710000003, 9853293.623709997],
                ['Q2-2004', 14300216.43, 10277650.763958748],
                ['Q3-2004', 14502421.63, 10466934.096533755],
                ['Q4-2004', 14321501.940000001, 10333095.839474997]]
        },

    :param raw_cellset_as_dict: raw data from TM1
    :param value_precision: Integer (optional) specifying number of decimal places to return
    :return: dict : { titles: [], headers: [axis][], cells: { Page0: [  [column name, column values], [], ... ], ...} }
    """
    header_map = build_headers_from_cellset(raw_cellset_as_dict, force_header_dimensionality=3)
    titles = header_map['titles']
    headers = header_map['headers']
    cardinality = header_map['cardinality']
    value_format_string = ""

    if value_precision:
        value_format_string = "{{0:.{}f}}".format(value_precision)

    cells = {}
    for z in range(cardinality[2]):
        z_header = headers[2][z]['name']
        page = []
        for x in range(cardinality[0]):
            x_header = headers[0][x]['name']
            row = [x_header]
            for y in range(cardinality[1]):
                cell_addr = (x + cardinality[0] * y + cardinality[0] * cardinality[1] * z)
                raw_value = raw_cellset_as_dict['Cells'][cell_addr]['Value'] or 0
                if value_precision:
                    row.append(float(value_format_string.format(raw_value)))
                else:
                    row.append(raw_value)
            page.append(row)
        cells[z_header] = page

    return {'titles': titles, 'headers': headers, 'cells': cells}


def build_headers_from_cellset(raw_cellset_as_dict: Dict, force_header_dimensionality: int = 1) -> Dict:
    """ Extract dimension headers from cellset into dictionary of titles (slicers) and headers (row,column,page)
    * Title dimensions are in a single list of dicts
    * Header dimensions are a 2-dimensional list of the element dicts
      * The first dimension in the header list is the axis
      * The second dimension is the list of elements on the axis
    * Dict format: {'name': 'element or compound name', 'members': [ {dict of dimension properties}, ... ] }
      * Stacked headers on an axis will have a compount 'name' created by joining the member's 'Name' properties with a '/'
      * Stacked headers will each be listed in the 'memebers' list; Single-element headers will only have one element in list
    :param raw_cellset_as_dict: raw data from TM1
    :param force_header_dimensionality: An optional integer (1,2 or 3) to force headers array to be at least that long
    :return: dict : { titles: [ { 'name': 'xx', 'members': {} } ], headers: [axis][ { 'name': 'xx', 'members': {} } ] }
    """
    dimensionality = len(raw_cellset_as_dict['Axes'])
    cardinality = [raw_cellset_as_dict['Axes'][axis]['Cardinality'] for axis in range(dimensionality)]

    titles = []
    headers = []
    for axis in range(dimensionality):
        members = []
        for tindex in range(cardinality[axis]):
            tuples_as_dict = raw_cellset_as_dict['Axes'][axis]['Tuples'][tindex]['Members']
            name = ' / '.join(tuple(member['Name'] for member in tuples_as_dict))
            members.append({'name': name, 'members': tuples_as_dict})

        if axis == dimensionality - 1 and cardinality[axis] == 1:
            titles = members
        else:
            headers.append(members)

    dimensionality = len(headers)
    cardinality = [len(headers[axis]) for axis in range(dimensionality)]

    # Handle 1, 2 and 3-dimensional cellsets. Use dummy row/page headers when missing
    if dimensionality == 1 and force_header_dimensionality > 1:
        headers += [[{'name': 'Row'}]]
        cardinality.insert(1, 1)
        dimensionality += 1
    if dimensionality == 2 and force_header_dimensionality > 2:
        headers += [[{'name': 'Page'}]]
        cardinality.insert(2, 1)
        dimensionality += 1

    return {'titles': titles, 'headers': headers, 'dimensionality': dimensionality, 'cardinality': cardinality}


def dimension_hierarchy_element_tuple_from_unique_name(element_unique_name: str) -> Tuple[str, str, str]:
    """ Extract dimension name, hierarchy name and element name from element unique name.
    Works with explicit and implicit hierarchy references.
    :param element_unique_name: e.g. [d1].[e1] or [d1].[leaves].[e1]
    :return: tuple of dimension name, hierarchy name, element name
    """
    dimension = dimension_name_from_element_unique_name(element_unique_name)
    element = element_name_from_element_unique_name(element_unique_name)
    if element_unique_name.count("].[") == 1:
        return dimension, dimension, element
    hierarchy = hierarchy_name_from_element_unique_name(element_unique_name)
    return dimension, hierarchy, element


def dimension_name_from_element_unique_name(element_unique_name: str) -> str:
    return element_unique_name[1:element_unique_name.find('].[')]


def hierarchy_name_from_element_unique_name(element_unique_name: str) -> str:
    return element_unique_name[element_unique_name.find('].[') + 3:element_unique_name.rfind('].[')]


def element_name_from_element_unique_name(element_unique_name: str) -> str:
    return element_unique_name[element_unique_name.rfind('].[') + 3:-1]


def element_names_from_element_unique_names(element_unique_names: Iterable[str]) -> Tuple[str]:
    """ Get tuple of simple element names from the full element unique names
    :param element_unique_names: tuple of element unique names ([dim1].[hier1].[elem1], ... )
    :return: tuple of element names: (elem1, elem2, ... )
    """
    return tuple(element_name_from_element_unique_name(unique_name)
                 for unique_name
                 in element_unique_names)


def dimension_names_from_element_unique_names(element_unique_names: Iterable[str]) -> Tuple[str]:
    """ Get tuple of simple element names from the full element unique names
    :param element_unique_names: tuple of element unique names ([dim1].[hier1].[elem1], ... )
    :return: tuple of element names: (elem1, elem2, ... )
    """
    return tuple(dimension_name_from_element_unique_name(unique_name)
                 for unique_name
                 in element_unique_names)


def build_element_unique_names(
        dimension_names: Iterable[str],
        element_names: Iterable[str],
        hierarchy_names: Optional[Iterable[str]] = None) -> Generator:
    """ Create tuple of unique names from dimension, hierarchy and elements

    :param dimension_names:
    :param element_names:
    :param hierarchy_names:
    :return: Generator
    """
    if not hierarchy_names:
        return ("[{}].[{}]".format(dim, elem)
                for dim, elem
                in zip(dimension_names, element_names))
    else:
        return ("[{}].[{}].[{}]".format(dim, hier, elem)
                for dim, hier, elem
                in zip(dimension_names, hierarchy_names, element_names))


@require_pandas
def build_pandas_dataframe_from_cellset(cellset: Dict, multiindex: bool = True,
                                        sort_values: bool = True) -> 'pd.DataFrame':
    """

    :param cellset:
    :param multiindex: True or False
    :param sort_values: Boolean to control sorting in result DataFrame
    :return:
    """
    try:
        cellset_clean = {}
        coordinates = []
        for coordinates, cell in cellset.items():
            element_names = element_names_from_element_unique_names(coordinates)
            cellset_clean[element_names] = cell['Value'] if cell else None
        dimension_names = tuple(unique_name[1:unique_name.find('].[')] for unique_name in coordinates)

        # create index
        keylist = list(cellset_clean.keys())
        index = pd.MultiIndex.from_tuples(keylist, names=dimension_names)

        # create DataFrame
        values = list(cellset_clean.values())
        df = pd.DataFrame(values, index=index, columns=["Values"])

        if not multiindex:
            df.reset_index(inplace=True)
            if sort_values:
                df.sort_values(inplace=True, by=list(dimension_names))
        return df
    except UnboundLocalError:
        message = """
            Can't build DataFrame from empty cellset. 
            Make sure the underlying MDX / View is not fully zero suppressed.
        """
        raise ValueError(message)


@require_pandas
def build_cellset_from_pandas_dataframe(
        df: 'pd.DataFrame',
        sum_numeric_duplicates: bool = True) -> 'CaseAndSpaceInsensitiveTuplesDict':
    """

    param sum_numeric_duplicates: Aggregate numerical values for duplicated intersections
    param df: a Pandas Dataframe, with dimension-column mapping in correct order.
    As created in build_pandas_dataframe_from_cellset

    :return: a CaseAndSpaceInsensitiveTuplesDict
    """
    if isinstance(df.index, pd.MultiIndex):
        df.reset_index(inplace=True)

    if sum_numeric_duplicates:
        value_header = df.columns[-1]
        dimension_headers = df.columns[:-1]

        if pd.api.types.is_numeric_dtype(df[value_header]):
            df = aggregate_duplicate_intersections(df, dimension_headers, value_header)
        else:
            filter_mask = df[value_header].apply(np.isreal)
            df_n = df[filter_mask]
            df_s = df[~ filter_mask]
            df_n = aggregate_duplicate_intersections(df_n, dimension_headers, value_header)
            df = pd.concat([df_n, df_s])

    cellset = CaseAndSpaceInsensitiveTuplesDict(
        dict(zip(df.iloc[:, :-1].itertuples(index=False, name=None), df.iloc[:, -1].values)))
    return cellset


def aggregate_duplicate_intersections(df, dimension_headers, value_header):
    return df.groupby([*dimension_headers])[value_header].sum().reset_index()


def lower_and_drop_spaces(item: str) -> str:
    return item.replace(" ", "").lower()


def get_seconds_from_duration(time_str: str) -> int:
    """
    This function will convert the TM1 time to seconds
    :param time_str: P0DT00H01M43S
    :return: int
    """
    import re
    pattern = re.compile('\w(\d+)\w\w(\d+)\w(\d+)\w(\d+)\w')
    matches = pattern.search(time_str)
    d, h, m, s = matches.groups()
    seconds = (int(d) * 86400) + (int(h) * 3600) + (int(m) * 60) + int(s)
    return seconds


def get_tm1_time_value_now(use_excel_serial_date: bool = False) -> float:
    """
    This function can be used to replicate TM1's NOW function
    to return current date/time value in serial number format.
    :param use_excel_serial_date: Boolean
    :return: serial number
    """
    from datetime import datetime
    # timestamp according to tm1
    start_datetime = datetime(1899, 12, 30) if use_excel_serial_date else datetime(1960, 1, 1)
    current_datetime = datetime.now()
    delta = current_datetime - start_datetime
    return delta.days + (delta.seconds / 86400)


def add_url_parameters(url, **kwargs: str) -> str:
    """ Append parameters to url string passed in kwargs
    :param url: str
    :param kwargs: key:value pairs of url parameters. For example, {'$select':'Name'}
    :return: str
    """
    parameters = []
    for key, value in kwargs.items():
        if value is not None:
            value = value.replace("'", "''") if isinstance(value, str) else value
            parameters.append(key + "=" + value)

    url_parts = list(urlparse.urlparse(url))
    query_part = url_parts[4]
    if query_part:
        query_part += "&"
    query_part += "&".join(parameters)

    url_parts[4] = query_part
    return urlparse.urlunparse(url_parts)


def extract_compact_json_cellset(context: str, response: Dict, return_as_dict: bool) -> Union[Dict, List]:
    """ Translates odata compact response json into default dictionary response or plain list (e.g., list of values)

    :param context: The context field from the TM1 response JSON
    :param response: The JSON response
    :param return_as_dict: boolean
    :return:
    """
    props = extract_cell_properties_from_odata_context(context)

    # First element [0] is the cellset ID, second is the cellset data
    cells_data = response['value'][1]

    # return props with data if required
    if return_as_dict:
        return map_cell_properties_to_compact_json_response(props, cells_data)

    if len(props) == 1:
        return [value[0] for value in cells_data]

    return cells_data


def extract_cell_properties_from_odata_context(context: str) -> List[str]:
    """ Takes in an odata_context and returns a list of properties e.g
      [Ordinal, Value, RuleDerived, ...]
    :param context: A valid odata_context returned when querying cells
    :return:
    """
    pattern = re.compile('\$metadata#Cellsets\(Cells\(([A-Za-z,]+)\)\)/\$entity')
    matches = pattern.match(context)
    if not matches:
        raise ValueError('Could not extract cell properties from odata context')
    cell_properties = matches.groups()[0].split(',')
    return cell_properties


def map_cell_properties_to_compact_json_response(properties: List, compact_cells_response: List) -> Dict:
    """ Map cell properties to compact json response e.g
    properties = [Ordinal, Value, RuleDerived]
    compact_cells_response = [[0, 258, 100], [1, 258, 500]]
    result: {Cells: [
        { Ordinal: 0, Value: 100, RuleDerived: 258}, 
        { Ordinal: 1, Value: 500, RuleDerived: 258}
    ]}
    

    :param properties: list of `Cell` properties e.g [Ordinal, Value, Updateable, ...]
    :param compact_cells_response: list of cells returned in compact json format
    :return: dict with properties mapped to compact json response    
    """
    cells_dict = dict()
    cells = []
    for cell in compact_cells_response:
        d = dict()
        for index, prop in enumerate(properties):
            d[prop] = cell[index]
        cells.append(d)
    cells_dict['Cells'] = cells
    return cells_dict


class CaseAndSpaceInsensitiveDict(collections.abc.MutableMapping):
    """A case-and-space-insensitive dict-like object with String keys.
    Implements all methods and operations of
    ``collections.abc.MutableMapping`` as well as dict's ``copy``. Also
    provides ``adjusted_items``, ``adjusted_keys``.
    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys.
    However, querying and contains testing is case insensitive:
        elements = TM1pyElementsDictionary()
        elements['Travel Expenses'] = 100
        elements['travelexpenses'] == 100 # True
    Entries are ordered
    """

    def __init__(self, data=None, **kwargs):
        self._store = collections.OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key: str, value):
        # Use the adjusted cased key for lookups, but store the actual
        # key alongside the value.
        self._store[lower_and_drop_spaces(key)] = (key, value)

    def __getitem__(self, key: str):
        return self._store[lower_and_drop_spaces(key)][1]

    def __delitem__(self, key: str):
        del self._store[lower_and_drop_spaces(key)]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self):
        return len(self._store)

    def adjusted_items(self) -> Generator:
        """Like iteritems(), but with all adjusted keys."""
        return (
            (adjusted_key, key_value[1])
            for (adjusted_key, key_value)
            in self._store.items()
        )

    def adjusted_keys(self) -> Generator:
        """Like keys(), but with all adjusted keys."""
        return (
            adjusted_key
            for (adjusted_key, key_value)
            in self._store.items()
        )

    def __eq__(self, other):
        if isinstance(other, collections.abc.Mapping):
            other = CaseAndSpaceInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.adjusted_items()) == dict(other.adjusted_items())

    # Copy is required
    def copy(self):
        return CaseAndSpaceInsensitiveDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))


class CaseAndSpaceInsensitiveTuplesDict(collections.abc.MutableMapping):
    """A case-and-space-insensitive dict-like object with String-Tuples Keys.
    Implements all methods and operations of
    ``collections.abc.MutableMapping`` as well as dict's ``copy``. Also
    provides ``adjusted_items``, ``adjusted_keys``.
    All keys are expected to be tuples of strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys.
    However, querying and contains testing is case insensitive:
        data = CaseAndSpaceInsensitiveTuplesDict()
        data[('[Business Unit].[UK]', '[Scenario].[Worst Case]')] = 1000
        data[('[BusinessUnit].[UK]', '[Scenario].[worstcase]')] == 1000 # True
        data[('[Business Unit].[UK]', '[Scenario].[Worst Case]')] == 1000 # True
    Entries are ordered
    """

    def __init__(self, data=None, **kwargs):
        self._store = collections.OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        # Use the adjusted cased key for lookups, but store the actual
        # key alongside the value.
        self._store[tuple([lower_and_drop_spaces(item) for item in key])] = (key, value)

    def __getitem__(self, key):
        return self._store[tuple([lower_and_drop_spaces(item) for item in key])][1]

    def __delitem__(self, key):
        del self._store[tuple([lower_and_drop_spaces(item) for item in key])]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self):
        return len(self._store)

    def items(self):
        return super(CaseAndSpaceInsensitiveTuplesDict, self).items()

    def adjusted_items(self):
        """Like iteritems(), but with all adjusted keys."""
        return (
            (adjusted_key, key_value[1])
            for (adjusted_key, key_value)
            in self._store.items()
        )

    def adjusted_keys(self):
        """Like keys(), but with all adjusted keys."""
        return (
            adjusted_key
            for (adjusted_key, key_value)
            in self._store.items()
        )

    def __eq__(self, other):
        if isinstance(other, collections.abc.Mapping):
            other = CaseAndSpaceInsensitiveTuplesDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.adjusted_items()) == dict(other.adjusted_items())

    # Copy is required
    def copy(self):
        return CaseAndSpaceInsensitiveTuplesDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))


class CaseAndSpaceInsensitiveSet(collections.abc.MutableSet):
    def __init__(self, *values):
        self._store = {}
        for v in values:
            if isinstance(v, str):
                self.add(v)
            elif isinstance(v, Iterable):
                for item in v:
                    self.add(item)
            else:
                self.add(v)

    def __contains__(self, value):
        return value.lower().replace(" ", "") in self._store

    def __delitem__(self, key):
        del self._store[key.lower().replace(" ", "")]

    def __iter__(self):
        return iter(self._store.values())

    def __len__(self):
        return len(self._store)

    def add(self, value):
        self._store[value.lower().replace(" ", "")] = value

    def discard(self, value):
        with suppress(KeyError):
            del self._store[value.lower().replace(" ", "")]

    def copy(self):
        return CaseAndSpaceInsensitiveSet(*self._store.values())

    def __repr__(self):
        return str(self._store)

    def __eq__(self, other):
        if isinstance(other, collections.abc.MutableSet):
            other = CaseAndSpaceInsensitiveSet(*other)
        else:
            return NotImplemented
        # Compare insensitively
        return set(self._store.keys()) == set(other._store.keys())

    def __sub__(self, other):
        result = self.copy()
        for entry in other:
            result.discard(entry)
        return result


def get_dimensions_from_where_clause(mdx: str) -> List[str]:
    mdx = mdx.replace(" ", "").upper()
    if "WHERE(" not in mdx:
        return []

    where = mdx[mdx.rfind("WHERE(") + 6:-1]
    unique_names = where.split(",")
    return [dimension_name_from_element_unique_name(unique_name) for unique_name in unique_names]


def get_cube(mdx: str) -> str:
    # replace tabs, line breaks, spaces
    mdx = re.sub(r'\s+', '', mdx)

    # happy case: cube name in square brackets
    pattern = r"(?s)(?i)FROM\[(.*?)\]"
    search_result = re.search(pattern, mdx)
    if search_result:
        return search_result.group(1)

    # cut off where
    pattern = r"(?s)(?i).*SELECT.*ON.*FROM.*WHERE\(.*"
    if re.search(pattern=pattern, string=mdx):
        # part before where
        mdx = re.split(r"(?s)(?i)WHERE\(.*", mdx)[0]

    # part after from
    cube = re.split(r"(?s)(?i)FROM", mdx)[-1]

    return cube


def resembles_mdx(mdx: str) -> bool:
    pattern = r"(?s)(?i).*SELECT.*ON.*FROM.*"
    if re.search(pattern=pattern, string=mdx):
        return True
    return False


def wrap_in_curly_braces(expression: str) -> str:
    """ Put curly braces around a string
    :param expression:
    :return:
    """
    return "".join(["{" if not expression.startswith("{") else "",
                    expression,
                    "}" if not expression.endswith("}") else ""])


@unique
class CellUpdateableProperty(Enum):
    SECURITY_RESTRICTED = 1
    UPDATE_CUBE_APPLICABLE = 2
    RULE_IS_APPLIED = 3
    PICKLIST_EXISTS = 4
    SANDBOX_VALUE_IS_DIFFERENT_TO_BASE = 5
    NO_SPREADING_HOLD = 9
    LEAF_HOLD = 10
    CONSOLIDATION_SPREADING_HOLD = 11
    TEMPORARY_SPREADING_HOLD = 12
    CELL_IS_NOT_UPDATEABLE = 29


def extract_cell_updateable_property(decimal_value: int, cell_property: CellUpdateableProperty) -> bool:
    """ Function converts passed decimal (integer) value to binary
    and extracts specified (cell_property) bit counting from the right.
    It will return TRUE if bit is set, and FALSE if bit is not set
    Each cell has 'Updateable' property - a decimal value, which needs to be converted to binary to get information
    about the cell

    :param decimal_value: int Decimal number
    :param cell_property: CellUpdateableProperty enum property to extract from decimal value
    :return: bool

    """
    bit = (decimal_value & (1 << cell_property.value - 1)) != 0
    return bit


def cell_is_updateable(cell: dict) -> bool:
    """ Function checks if the cell can be updated
    :param cell: dict cell including Updateable property
    :return: bool
    """
    if "Updateable" not in cell:
        raise ValueError("cell dictionary must contain key 'Updateable'")

    bit = extract_cell_updateable_property(cell["Updateable"], CellUpdateableProperty.CELL_IS_NOT_UPDATEABLE)
    updateable = not bit
    return updateable


def build_mdx_from_cellset(cells: Dict, cube_name: str, dimensions: Iterable[str]) -> str:
    query = MdxBuilder.from_cube(cube_name)
    for coordinates in cells:
        members = (Member.of(dimension, element) for dimension, element in zip(dimensions, coordinates))
        query.add_member_tuple_to_columns(*members)
    mdx = query.to_mdx()
    return mdx


def build_mdx_and_values_from_cellset(cells: Dict, cube_name: str, dimensions: Iterable[str]) -> Tuple[str, List]:
    values = []
    query = MdxBuilder.from_cube(cube_name)
    for coordinates, value in cells.items():
        members = (Member.of(dimension, element) for dimension, element in zip(dimensions, coordinates))
        query.add_member_tuple_to_columns(*members)
        values.append(value)
    mdx = query.to_mdx()
    return mdx, values


def frame_to_significant_digits(x, digits=15):
    if x == 0 or not math.isfinite(x):
        return str(x).replace('e+', 'E')
    digits -= math.ceil(math.log10(abs(x)))
    return str(round(x, digits)).replace('e+', 'E')


class HTTPAdapterWithSocketOptions(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.socket_options = kwargs.pop("socket_options", None)
        super(HTTPAdapterWithSocketOptions, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        if self.socket_options is not None:
            kwargs["socket_options"] = self.socket_options
        super(HTTPAdapterWithSocketOptions, self).init_poolmanager(*args, **kwargs)



# -*- coding: utf-8 -*-

import collections
import json
from typing import Dict, Union

from TM1py.Objects.Subset import Subset, AnonymousSubset
from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import format_url


class ViewAxisSelection(TM1Object):
    """ Describes what is selected in a dimension on an axis. Can be a Registered Subset or an Anonymous Subset

    """

    def __init__(self, dimension_name: str, subset: Union[Subset, AnonymousSubset]):
        """
            :Parameters:
                `dimension_name` : String
                `subset` : Subset or AnonymousSubset
        """
        self._subset = subset
        self._dimension_name = dimension_name
        self._hierarchy_name = dimension_name

    @property
    def subset(self) -> Union[Subset, AnonymousSubset]:
        return self._subset

    @property
    def dimension_name(self) -> str:
        return self._dimension_name

    @property
    def hierarchy_name(self) -> str:
        return self._hierarchy_name

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body(), ensure_ascii=False)

    @property
    def body_as_dict(self) -> Dict:
        return self._construct_body()

    def _construct_body(self) -> Dict:
        """ construct the ODATA conform JSON represenation for the ViewAxisSelection entity.

        :return: dictionary
        """
        body_as_dict = collections.OrderedDict()
        if isinstance(self._subset, AnonymousSubset):
            body_as_dict['Subset'] = json.loads(self._subset.body)
        elif isinstance(self._subset, Subset):
            subset_path = format_url(
                "Dimensions('{}')/Hierarchies('{}')/Subsets('{}')",
                self._dimension_name, self._hierarchy_name, self._subset.name)
            body_as_dict['Subset@odata.bind'] = subset_path
        return body_as_dict


class ViewTitleSelection:
    """ Describes what is selected in a dimension on the view title.
        Can be a Registered Subset or an Anonymous Subset

    """

    def __init__(self, dimension_name: str, subset: Union[AnonymousSubset, Subset], selected: str):
        self._dimension_name = dimension_name
        self._hierarchy_name = dimension_name
        self._subset = subset
        self._selected = selected

    @property
    def subset(self) -> Union[Subset, AnonymousSubset]:
        return self._subset

    @property
    def dimension_name(self) -> str:
        return self._dimension_name

    @property
    def hierarchy_name(self) -> str:
        return self._hierarchy_name

    @property
    def selected(self) -> str:
        return self._selected

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body(), ensure_ascii=False)

    def _construct_body(self) -> Dict:
        """ construct the ODATA conform JSON represenation for the ViewTitleSelection entity.

        :return: string, the valid JSON
        """
        body_as_dict = collections.OrderedDict()
        if isinstance(self._subset, AnonymousSubset):
            body_as_dict['Subset'] = json.loads(self._subset.body)
        elif isinstance(self._subset, Subset):
            subset_path = format_url(
                "Dimensions('{}')/Hierarchies('{}')/Subsets('{}')",
                self._dimension_name, self._hierarchy_name, self._subset.name)
            body_as_dict['Subset@odata.bind'] = subset_path
        element_path = format_url(
            "Dimensions('{}')/Hierarchies('{}')/Elements('{}')",
            self._dimension_name, self._hierarchy_name, self._selected)
        body_as_dict['Selected@odata.bind'] = element_path
        return body_as_dict



# -*- coding: utf-8 -*-

import collections
import json
from typing import Dict

from TM1py.Objects.TM1Object import TM1Object


class Sandbox(TM1Object):
    """ Abstraction of a TM1 Sandbox
        
    """

    def __init__(self, name: str, include_in_sandbox_dimension: bool = True):
        """
        
        :param name: name of the Sandbox
        :param include_in_sandbox_dimension: 
        """
        self.name = name
        self.include_in_sandbox_dimension = include_in_sandbox_dimension

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def include_in_sandbox_dimension(self) -> bool:
        return self._include_in_sandbox_dimension

    @include_in_sandbox_dimension.setter
    def include_in_sandbox_dimension(self, value: bool):
        self._include_in_sandbox_dimension = value

    @classmethod
    def from_json(cls, sandbox_as_json: str) -> "Sandbox":
        """ Alternative constructor

        :param sandbox_as_json: user as JSON string
        :return: sandbox, an instance of this class
        """
        sandbox_as_dict = json.loads(sandbox_as_json)
        return cls.from_dict(sandbox_as_dict)

    @classmethod
    def from_dict(cls, sandbox_as_dict: Dict) -> "Sandbox":
        """ Alternative constructor

        :param sandbox_as_dict: user as dict
        :return: an instance of this class
        """
        return cls(
            name=sandbox_as_dict["Name"],
            include_in_sandbox_dimension=sandbox_as_dict["IncludeInSandboxDimension"],
        )

    @property
    def body(self) -> str:
        return self._construct_body()

    def _construct_body(self) -> str:
        """
        construct body (json) from the class attributes
        :return: String, TM1 JSON representation of a sandbox
        """
        body_as_dict = collections.OrderedDict()
        body_as_dict["Name"] = self.name
        body_as_dict["IncludeInSandboxDimension"] = self._include_in_sandbox_dimension
        return json.dumps(body_as_dict, ensure_ascii=False)



# -*- coding: utf-8 -*-

import collections
import json
from enum import Enum
from typing import Iterable, Optional, List, Dict, Union

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils.Utils import CaseAndSpaceInsensitiveSet, format_url


class UserType(Enum):
    User = 0
    SecurityAdmin = 1
    DataAdmin = 2
    Admin = 3
    OperationsAdmin = 4

    def __str__(self):
        return self.name

    @classmethod
    def _missing_(cls, value: str):
        for member in cls:
            if member.name.lower() == value.replace(" ", "").lower():
                return member
        # default
        raise ValueError("Invalid element type=" + value)


class User(TM1Object):
    """ Abstraction of a TM1 User
    
    """

    def __init__(self, name: str, groups: Iterable[str], friendly_name: Optional[str] = None,
                 password: Optional[str] = None, user_type: Union[UserType, str] = None, enabled: bool = None):
        self._name = name
        self._groups = CaseAndSpaceInsensitiveSet(*groups)
        self._friendly_name = friendly_name
        self._password = password
        self._enabled = enabled
        self._user_type = user_type
        # determine user_type
        if user_type is None:
            if str(UserType.Admin) in self._groups:
                self.user_type = UserType.Admin
            elif str(UserType.SecurityAdmin) in self._groups:
                self.user_type = UserType.SecurityAdmin
            elif str(UserType.DataAdmin) in self._groups:
                self.user_type = UserType.DataAdmin
            elif str(UserType.OperationsAdmin) in self._groups:
                self.user_type = UserType.OperationsAdmin
            else:
                self.user_type = UserType.User
        else:
            self.user_type = user_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def user_type(self) -> UserType:
        return self._user_type

    @property
    def friendly_name(self) -> str:
        return self._friendly_name

    @property
    def password(self) -> str:
        if self._password:
            return self._password

    @property
    def is_admin(self) -> bool:
        return "ADMIN" in CaseAndSpaceInsensitiveSet(*self.groups)

    @property
    def groups(self) -> List[str]:
        return [group for group in self._groups]

    @property
    def enabled(self) -> bool:
        return self._enabled

    @name.setter
    def name(self, value: str):
        self._name = value

    @friendly_name.setter
    def friendly_name(self, value: str):
        self._friendly_name = value

    @password.setter
    def password(self, value: str):
        self._password = value

    @enabled.setter
    def enabled(self, value: Union[bool, None]):
        self._enabled = value

    @user_type.setter
    def user_type(self, value: Union[str, UserType]):
        if not isinstance(value, str) and not isinstance(value, UserType):
            raise ValueError("argument 'user_type' must be of type str or UserType")

        self._user_type = UserType(value)
        # update groups as well, since TM1 doesn't react to change in user_type property
        if self._user_type is not UserType.User:
            self.add_group(str(self._user_type))

    def add_group(self, group_name: str):
        self._groups.add(group_name)

    def remove_group(self, group_name: str):
        self._groups.discard(group_name)

    @classmethod
    def from_json(cls, user_as_json: str):
        """ Alternative constructor

        :param user_as_json: user as JSON string
        :return: user, an instance of this class
        """
        user_as_dict = json.loads(user_as_json)
        return cls.from_dict(user_as_dict)

    @classmethod
    def from_dict(cls, user_as_dict: Dict) -> 'User':
        """ Alternative constructor

        :param user_as_dict: user as dict
        :return: user, an instance of this class
        """
        return cls(name=user_as_dict['Name'],
                   friendly_name=user_as_dict['FriendlyName'],
                   enabled=user_as_dict["Enabled"],
                   user_type=user_as_dict["Type"],
                   groups=[group["Name"] for group in user_as_dict['Groups']])

    @property
    def body(self) -> str:
        return self.construct_body()

    def construct_body(self) -> str:
        """
        construct body (json) from the class attributes
        :return: String, TM1 JSON representation of a user
        """
        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self.name
        body_as_dict['FriendlyName'] = self.friendly_name or self.name
        body_as_dict['Enabled'] = self._enabled
        body_as_dict['Type'] = str(self._user_type)
        if self.password:
            body_as_dict['Password'] = self._password
        body_as_dict['Groups@odata.bind'] = [format_url("Groups('{}')", group)
                                             for group
                                             in self.groups]
        return json.dumps(body_as_dict, ensure_ascii=False)



# -*- coding: utf-8 -*-

import datetime


class ChoreStartTime:
    """ Utility class to handle time representation for Chore Start Time
        
    """

    def __init__(self, year: int, month: int, day: int, hour: int, minute: int, second: int, tz: str = None):
        """
        
        :param year: year 
        :param month: month
        :param day: day
        :param hour: hour or None
        :param minute: minute or None
        :param second: second or None
        """
        self._datetime = datetime.datetime.combine(datetime.date(year, month, day), datetime.time(hour, minute, second))
        self.tz = tz

    @classmethod
    def from_string(cls, start_time_string: str) -> 'ChoreStartTime':
        # extract optional tz info (e.g., +01:00) from string end
        if '+' in start_time_string:
            # case "2020-11-05T08:00:01+01:00",
            tz = "+" + start_time_string.split('+')[1]
        elif start_time_string.count('-') == 3:
            # case: "2020-11-05T08:00:01-01:00",
            tz = "-" + start_time_string.split('-')[-1]
        else:
            tz = None

        # f to handle strange timestamp 2016-09-25T20:25Z instead of common 2016-09-25T20:25:00Z
        f = lambda x: int(x) if x else 0
        return cls(year=f(start_time_string[0:4]),
                   month=f(start_time_string[5:7]),
                   day=f(start_time_string[8:10]),
                   hour=f(start_time_string[11:13]),
                   minute=f(start_time_string[14:16]),
                   second=f(start_time_string[17:19]),
                   tz=tz)

    @property
    def start_time_string(self) -> str:
        # produce timestamp 2016-09-25T20:25Z instead of common 2016-09-25T20:25:00Z
        if not self._datetime.second:
            start_time = self._datetime.strftime("%Y-%m-%dT%H:%M")
        else:
            start_time = self._datetime.strftime("%Y-%m-%dT%H:%M:%S")

        if self.tz:
            start_time += self.tz
        else:
            start_time += "Z"

        return start_time

    @property
    def datetime(self) -> datetime:
        return self._datetime

    def __str__(self):
        return self.start_time_string

    def set_time(self, year: int = None, month: int = None, day: int = None, hour: int = None, minute: int = None,
                 second: int = None):

        _year = year if year is not None else self._datetime.year
        _month = month if month is not None else self._datetime.month
        _day = day if day is not None else self._datetime.day
        _hour = hour if hour is not None else self._datetime.hour
        _minute = minute if minute is not None else self._datetime.minute
        _second = second if second is not None else self._datetime.second

        self._datetime = self._datetime.replace(year=_year,month=_month,day=_day,hour=_hour,minute=_minute,second=_second)

    def add(self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0):
        self._datetime = self._datetime + datetime.timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)

    def subtract(self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0):
        self._datetime = self._datetime - datetime.timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)



# -*- coding: utf-8 -*-

import collections
import json
from typing import Iterable, Dict, List

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import format_url


class Annotation(TM1Object):
    """ Abtraction of TM1 Annotation

        :Notes:
            - Class complete, functional and tested.
            - doesn't cover Attachments though
    """

    def __init__(self, comment_value: str, object_name: str, dimensional_context: Iterable[str],
                 comment_type: str = 'ANNOTATION', annotation_id: str = None,
                 text: str = '', creator: str = None, created: str = None, last_updated_by: str = None,
                 last_updated: str = None):
        self._id = annotation_id
        self._text = text
        self._creator = creator
        self._created = created
        self._last_updated_by = last_updated_by
        self._last_updated = last_updated
        self._dimensional_context = list(dimensional_context)
        self._comment_type = comment_type
        self._comment_value = comment_value
        self._object_name = object_name

    @classmethod
    def from_json(cls, annotation_as_json: str) -> 'Annotation':
        """ Alternative constructor

            :param annotation_as_json: String, JSON
            :return: instance of TM1py.Process
        """
        annotation_as_dict = json.loads(annotation_as_json)
        annotation_id = annotation_as_dict['ID']
        text = annotation_as_dict['Text']
        creator = annotation_as_dict['Creator']
        created = annotation_as_dict['Created']
        last_updated_by = annotation_as_dict['LastUpdatedBy']
        last_updated = annotation_as_dict['LastUpdated']
        dimensional_context = [item['Name'] for item in annotation_as_dict['DimensionalContext']]
        comment_type = annotation_as_dict['commentType']
        comment_value = annotation_as_dict['commentValue']
        object_name = annotation_as_dict['objectName']
        return cls(comment_value=comment_value, object_name=object_name, dimensional_context=dimensional_context,
                   comment_type=comment_type, annotation_id=annotation_id, text=text, creator=creator, created=created,
                   last_updated_by=last_updated_by, last_updated=last_updated)

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body())

    @property
    def body_as_dict(self) -> Dict:
        return self._construct_body()

    @property
    def comment_value(self) -> str:
        return self._comment_value

    @property
    def text(self) -> str:
        return self._text

    @property
    def dimensional_context(self) -> List[str]:
        return self._dimensional_context

    @property
    def created(self) -> str:
        return self._created

    @property
    def object_name(self) -> str:
        return self._object_name

    @property
    def last_updated(self) -> str:
        return self._last_updated

    @property
    def last_updated_by(self) -> str:
        return self._last_updated_by

    @comment_value.setter
    def comment_value(self, value: str):
        self._comment_value = value

    @property
    def id(self) -> str:
        return self._id

    def move(self, dimension_order: Iterable[str], dimension: str, target_element: str, source_element: str = None):
        """ Move annotation on given dimension from source_element to target_element
        
            :param dimension_order: List, order of the dimensions in the cube
            :param dimension: dimension name
            :param target_element: target element name
            :param source_element:  source element name
            :return: 
        """
        for i, dimension_name in enumerate(dimension_order):
            if dimension_name.lower() == dimension.lower():
                if not source_element or self._dimensional_context[i] == source_element:
                    self._dimensional_context[i] = target_element

    def _construct_body(self) -> Dict:
        """ construct the ODATA conform JSON represenation for the Annotation entity.

            :return: string, the valid JSON
        """
        dimensional_context = [{'Name': element} for element in self._dimensional_context]
        body = collections.OrderedDict()
        body['ID'] = self._id
        body['Text'] = self._text
        body['Creator'] = self._creator
        body['Created'] = self._created
        body['LastUpdatedBy'] = self._last_updated_by
        body['LastUpdated'] = self._last_updated
        body['DimensionalContext'] = dimensional_context
        comment_locations = ','.join(self._dimensional_context)
        body['commentLocation'] = comment_locations[1:]
        body['commentType'] = self._comment_type
        body['commentValue'] = self._comment_value
        body['objectName'] = self._object_name
        return body

    def construct_body_for_post(self, cube_dimensions) -> Dict:
        body = collections.OrderedDict()
        body["Text"] = self.text
        body["ApplicationContext"] = [{
            "Facet@odata.bind": "ApplicationContextFacets('}Cubes')",
            "Value": self.object_name}]
        body["DimensionalContext@odata.bind"] = []

        for dimension, element in zip(cube_dimensions, self.dimensional_context):
            coordinates = format_url("Dimensions('{}')/Hierarchies('{}')/Members('{}')", dimension, dimension, element)
            body["DimensionalContext@odata.bind"].append(coordinates)

        body['objectName'] = self.object_name
        body['commentValue'] = self.comment_value
        body['commentType'] = 'ANNOTATION'
        body['commentLocation'] = ','.join(self.dimensional_context)

        return body
